import json
import re
from typing import Dict, Any, Tuple, List, Optional

from eth_utils import encode_hex, event_abi_to_log_topic, decode_hex
from pyspark.sql.types import StructType, LongType, DataType, StringType, DoubleType, BooleanType, ArrayType
from web3 import Web3

# ABI types: https://docs.soliditylang.org/en/v0.8.11/abi-spec.html#types
abi_to_spark_type: Dict[str, DataType] = {
    r'^uint[^\[\]]*$': LongType(),
    r'^int[^\[\]]*$': LongType(),
    r'^address$': StringType(),
    r'^bool$': BooleanType(),
    r'^fixed.*x[^\[\]]*$': DoubleType(),
    r'^ufixed.*x[^\[\]]*$': DoubleType(),
    r'^bytes[^\[\]]*$': StringType(),
    r'^string$': StringType()
}


def _get_spark_type_by_name(type_name: str) -> DataType:
    for regex_str, data_type in abi_to_spark_type.items():
        if re.search(regex_str, type_name) is not None:
            return data_type

        if type_name.endswith('[]'):
            return ArrayType(_get_spark_type_by_name(regex_str[:-2]))


def _flatten_schema_from_components(component_abi: [Dict[str, Any]]) -> StructType:
    struct_type = StructType()
    for field in component_abi:
        # the default name of filed is 'param'
        fname = field.get('name') if len(field.get('name', '')) > 0 else 'param'
        ftype = field.get('type')
        if ftype != 'tuple':
            struct_type.add(field=fname, data_type=_get_spark_type_by_name(ftype))
        else:
            struct_type.add(
                field=fname,
                data_type=_flatten_schema_from_components(component_abi=field.get('components'))
            )

    return struct_type


def get_call_schema(abi: [Dict[str, Any]], mode: str) -> StructType:
    assert (mode == 'function' or mode == 'event')

    struct_type = StructType()
    func_abi_list = [i for i in abi if i.get('type') == mode]

    for func_abi in func_abi_list:
        ftype = func_abi.get('inputs', [])
        if len(ftype) == 0:
            continue

        struct_type.add(
            field=func_abi.get('name'),
            data_type=_flatten_schema_from_components(component_abi=func_abi.get('inputs'))
        )

    return struct_type


def _flatten_tuple(raw_tuple: Tuple, schema: List[Dict[str, Any]]) -> Dict[str, Any]:
    result_dict = {}

    for index, field_type in enumerate(schema):
        fname = field_type.get('name')
        if field_type.get('type') != 'tuple':
            result_dict[fname] = raw_tuple[index]
        else:
            result_dict[fname] = _flatten_tuple(raw_tuple=raw_tuple[index],
                                                schema=field_type.get('components'))

    return result_dict


def _flatten_params(params: Dict[str, Any], schema: List[Dict[str, Any]]) -> Dict[str, Any]:
    result_dict = {}
    for key, value in params.items():
        fname = key if len(key) > 0 else 'param'

        if type(value) is not tuple:
            result_dict[fname] = value
        else:
            component_schema = [i for i in schema if i.get('name') == key and i.get('type') == 'tuple'][0] \
                .get('components')

            result_dict[fname] = _flatten_tuple(raw_tuple=value,
                                                schema=component_schema)

    return result_dict


def decode_function_input_with_abi(input_data: str, abi_str: str) -> Optional[Dict[str, Any]]:
    try:
        w3 = Web3()
        abi = json.loads(abi_str)
        contract = w3.eth.contract(abi=abi)
        func_obj, input_params = contract.decode_function_input(input_data)
        func_name = vars(func_obj)['fn_name']
        input_schema = [i for i in abi if i.get('name') == func_name and i.get('type') == 'function'][0].get('inputs')
        return {func_name: _flatten_params(params=input_params, schema=input_schema)}
    except Exception as ex:
        print(f'parsing input with abi failed: {str(ex)}')
        return None


def decode_log_data_with_abi(data: str, topics_str: str, abi_str: str) -> Optional[Dict[str, Any]]:
    try:
        w3 = Web3()
        abi = json.loads(abi_str)
        contract = w3.eth.contract(abi=abi)
        topics = topics_str.split(',')

        event_abi_list = [i for i in abi if 'name' in i and topics[0] == encode_hex(event_abi_to_log_topic(i))]

        if len(event_abi_list) == 0:
            raise Exception('Could not find any event with matching selector.')

        schema = event_abi_list[0].get('inputs')
        name = event_abi_list[0].get('name')

        event = contract.events[name]().processLog({
            'data': data,
            'topics': [decode_hex(topic) for topic in topics],
            # Placeholder
            'logIndex': '',
            'transactionIndex': '',
            'transactionHash': '',
            'address': '',
            'blockHash': '',
            'blockNumber': ''
        })

        return {name: _flatten_params(params=event.args, schema=schema)}
    except Exception as ex:
        print(f'parsing log data with abi failed: {str(ex)}')
        return None



import unittest

import pyspark


class PySparkTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        conf = pyspark.SparkConf() \
            .setMaster("local[*]") \
            .setAppName("testing") \
            .set("spark.driver.bindAddress", "127.0.0.1")

        cls.sc = pyspark.SparkContext(conf=conf)
        cls.spark = pyspark.SQLContext(cls.sc)

    @classmethod
    def tearDownClass(cls):
        cls.sc.stop()



import os

from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


long_description = read('README.md') if os.path.isfile("README.md") else ""

setup(
    name='blockchain-spark',
    version='0.1.0',
    author='songv',
    author_email='songwei@iftech.io',
    description='Spark extension utils for blockchain',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/tellery/blockchain-spark',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ],
    keywords='web3, pandas',
    python_requires='>=3.6,<4',
    install_requires=[
        'pyspark==3.2.1',
        'web3==5.26.0'
    ],
    project_urls={
        'Bug Reports': 'https://github.com/tellery/blockchain-spark/issues',
        'Source': 'https://github.com/tellery/blockchain-spark',
    },
)



import json
import unittest
from typing import AnyStr

from pyspark.sql.types import DataType, StructType

import test.resources
from spark3.ethereum import decode_function_input_with_abi, get_call_schema, decode_log_data_with_abi

RESOURCE_GROUP = 'ethereum_test'


def _get_resource_path(file_name: str) -> AnyStr:
    return test.resources.get_resource_path([RESOURCE_GROUP], file_name)


def _read_resource(file_name: str) -> AnyStr:
    return test.resources.read_resource([RESOURCE_GROUP], file_name)


def deep_get_data_type(path: str, dtype: StructType) -> DataType:
    split_path = path.split('.')
    item_type = dtype[split_path[0]].dataType
    return item_type if len(split_path) == 1 else \
        deep_get_data_type('.'.join(split_path[1:]), item_type)


class SchemaTestCase(unittest.TestCase):

    def test_get_function_call_schema(self):
        abi = json.loads(_read_resource('trace_abi1.json'))
        struct_type = get_call_schema(abi=abi, mode='function')

        # MINT_WITH_SIG_TYPEHASH has no inputs
        self.assertFalse('MINT_WITH_SIG_TYPEHASH' in struct_type.fieldNames())
        # PERMIT_TYPEHASH has no inputs
        self.assertFalse('PERMIT_TYPEHASH' in struct_type.fieldNames())
        self.assertTrue('getApproved' in struct_type.fieldNames())

        self.assertTrue(deep_get_data_type('acceptBid', struct_type).typeName(), 'struct')
        self.assertTrue(deep_get_data_type('acceptBid.tokenId', struct_type).typeName(), 'long')
        self.assertTrue(deep_get_data_type('acceptBid.bid', struct_type).typeName(), 'struct')
        self.assertTrue(deep_get_data_type('acceptBid.bid.amount', struct_type).typeName(), 'long')
        self.assertTrue(deep_get_data_type('acceptBid.bid.currency', struct_type).typeName(), 'string')

        # TODO test complex ABI type

    def test_decode_function_input_with_abi(self):
        abi = _read_resource('trace_abi1.json')
        data = _read_resource('trace_input1.txt')
        input_data = decode_function_input_with_abi(input_data=data, abi_str=abi)

        self.assertEqual(
            input_data['mint']['data']['tokenURI'],
            'https://ipfs.fleek.co/ipfs/bafybeifyqibqlheu7ij7fwdex4y2pw2wo7eaw2z6lec5zhbxu3cvxul6h4')
        self.assertEqual(
            input_data['mint']['data']['metadataHash'],
            b'\x9e\xf9\xd6\xd1\xdc<\xfb\xd36T\x95\x078l@36C0Puv_d-\xb4\x1f\xb5v\xf8\xf9\x9d')
        self.assertEqual(input_data['mint']['bidShares']['prevOwner']['value'], 0)
        self.assertEqual(input_data['mint']['bidShares']['owner']['value'], 100000000000000000000)

        # TODO test complex ABI type data

    def test_decode_log_data_with_abi(self):
        abi = _read_resource('log_abi1.json')
        log = json.loads(_read_resource('log1.json'))
        log_data = decode_log_data_with_abi(data=log.get('data'), topics_str=log.get('topics'), abi_str=abi)

        self.assertEqual(log_data['OrdersMatched']['maker'], '0x6Be4a7bbb812Bfa6A63126EE7b76C8A13529BDB8')
        self.assertEqual(log_data['OrdersMatched']['taker'], '0x0239769A1aDF4DeF9f07Da824B80B9C4fCB59593')
        self.assertEqual(
            log_data['OrdersMatched']['metadata'],
            b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
        self.assertEqual(
            log_data['OrdersMatched']['price'],
            2000000000000000
        )

        # TODO test complex ABI type data



import os
from typing import AnyStr


def get_resource_path(groups: [str], file_name: str) -> str:
    current_file_dir = os.path.dirname(__file__)
    fixture_file_name = os.path.join(current_file_dir, *groups, file_name)

    if not os.path.exists(fixture_file_name):
        raise ValueError('File does not exist: ' + fixture_file_name)

    return fixture_file_name


def read_resource(groups: [str], file_name: str) -> AnyStr:
    fixture_file_name = get_resource_path(groups, file_name)

    with open(fixture_file_name, encoding="utf-8") as file_handle:
        return file_handle.read()



import json

import pyspark.sql
from pyspark.sql.functions import udf, lit, col

from spark3.ethereum import decode_function_input_with_abi, get_call_schema, decode_log_data_with_abi


class Transformer:

    @staticmethod
    def parse_traces_input(df: pyspark.sql.DataFrame, abi: str) -> pyspark.sql.DataFrame:
        # The dataframe must container input column.
        assert (len([col_name for col_name, col_type in df.dtypes if
                     col_name == 'input' and col_type == 'string']) == 1)

        schema = get_call_schema(abi=json.loads(abi), mode='function')
        parse_traces_udf = udf(f=lambda input_data, abi_str: decode_function_input_with_abi(input_data, abi_str),
                               returnType=schema)

        return df.filter(col('input').isNotNull) \
            .withColumn('function_parameters', parse_traces_udf('input', lit(abi))) \
            .filter(col('function_parameters').isNotNull())

    @staticmethod
    def parse_logs_data(df: pyspark.sql.DataFrame, abi: str) -> pyspark.sql.DataFrame:
        # The dataframe must container topics and data column.
        assert (len([col_name for col_name, col_type in df.dtypes if
                     col_name == 'topics' and col_type == 'string']) == 1)

        assert (len([col_name for col_name, col_type in df.dtypes if
                     col_name == 'data' and col_type == 'string']) == 1)

        schema = get_call_schema(abi=json.loads(abi), mode='event')
        parse_logs_udf = udf(f=lambda data, topics, abi_str: decode_log_data_with_abi(data, topics, abi_str),
                             returnType=schema)

        return df.filter(col('data').isNotNull) \
            .withColumn('log_parameters', parse_logs_udf('data', 'topics', lit(abi))) \
            .filter(col('log_parameters').isNotNull())
